/*
 * Copyright (c) 2009 Lukas Mejdrech
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup netif
 *  @{
 */

/** @file
 */

#include <async.h>
#include <mem.h>
#include <rwlock.h>
#include <stdio.h>

#include <ipc/ipc.h>
#include <ipc/services.h>

#include "../err.h"
#include "../messages.h"
#include "../modules.h"

#include "../structures/packet/packet.h"
#include "../structures/packet/packet_client.h"
#include "../structures/measured_strings.h"

#include "device.h"
#include "netif.h"
#include "netif_interface.h"

#define IPC_GET_DEVICE( call )		( device_id_t ) IPC_GET_ARG1( * call )
#define IPC_GET_PACKET( call )		( packet_id_t ) IPC_GET_ARG2( * call )
#define IPC_GET_IRQ( call )		( int ) IPC_GET_ARG2( * call )
#define IPC_GET_IO( call )		( int ) IPC_GET_ARG3( * call )
#define IPC_GET_PHONE( call )		( int ) IPC_GET_ARG5( * call )

extern netif_globals_t netif_globals;

DEVICE_MAP_IMPLEMENT( device_map, device_t )

int	module_message( ipc_callid_t callid, ipc_call_t * call, ipc_call_t * answer, int * answer_count );
int	netif_start_module( async_client_conn_t client_connection );
int	register_message( device_id_t device_id, int phone );

int find_device( device_id_t device_id, device_ref * device ){
	if( ! device ) return EBADMEM;
	* device = device_map_find( & netif_globals.device_map, device_id );
	if( ! * device ) return ENOENT;
	if(( ** device ).state == NETIF_NULL ) return EPERM;
	return EOK;
}

void null_device_stats( device_stats_ref stats ){
	bzero( stats, sizeof( device_stats_t ));
}

int register_message( device_id_t device_id, int phone ){
	ERROR_DECLARE;

	device_ref	device;

	ERROR_PROPAGATE( find_device( device_id, & device ));
	if( device->nil_phone > 0 ) return ELIMIT;
	device->nil_phone = phone;
	printf( "\nNew receiver of the device %d registered:\n\tphone\t= %d", device->device_id, device->nil_phone );
	return EOK;
}

int netif_message( ipc_callid_t callid, ipc_call_t * call, ipc_call_t * answer, int * answer_count ){
	ERROR_DECLARE;

	size_t				length;
	device_stats_t		stats;
	packet_t			packet;
	measured_string_t	address;

	* answer_count = 0;
	switch( IPC_GET_METHOD( * call )){
		case IPC_M_PHONE_HUNGUP:
			return EOK;
		case NET_NETIF_PROBE_AUTO:
			rwlock_write_lock( & netif_globals.lock );
			ERROR_CODE = probe_auto_message();
			rwlock_write_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case NET_NETIF_PROBE:
			rwlock_write_lock( & netif_globals.lock );
			ERROR_CODE = probe_message( IPC_GET_DEVICE( call ), IPC_GET_IRQ( call ), IPC_GET_IO( call ));
			rwlock_write_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case IPC_M_CONNECT_TO_ME:
			rwlock_write_lock( & netif_globals.lock );
			ERROR_CODE = register_message( IPC_GET_DEVICE( call ), IPC_GET_PHONE( call ));
			rwlock_write_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case NET_NETIF_SEND:
			rwlock_write_lock( & netif_globals.lock );
			if( ! ERROR_OCCURRED( packet_translate( netif_globals.networking_phone, & packet, IPC_GET_PACKET( call )))){
				ERROR_CODE = send_message( IPC_GET_DEVICE( call ), packet );
			}
			rwlock_write_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case NET_NETIF_START:
			rwlock_write_lock( & netif_globals.lock );
			ERROR_CODE = start_message( IPC_GET_DEVICE( call ));
			rwlock_write_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case NET_NETIF_STATS:
			rwlock_read_lock( & netif_globals.lock );
			if( ! ERROR_OCCURRED( ipc_data_read_receive( & callid, & length ))){
				if( length < sizeof( device_stats_t )){
					ERROR_CODE = EOVERFLOW;
				}else{
					if( ! ERROR_OCCURRED( get_device_stats( IPC_GET_DEVICE( call ), & stats ))){
						ERROR_CODE = ipc_data_read_finalize( callid, & stats, sizeof( device_stats_t ));
					}
				}
			}
			rwlock_read_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case NET_NETIF_STOP:
			rwlock_write_lock( & netif_globals.lock );
			ERROR_CODE = stop_message( IPC_GET_DEVICE( call ));
			rwlock_write_unlock( & netif_globals.lock );
			return ERROR_CODE;
		case NET_NETIF_GET_ADDR:
			rwlock_read_lock( & netif_globals.lock );
			if( ! ERROR_OCCURRED( get_addr_message( IPC_GET_DEVICE( call ), & address ))){
				ERROR_CODE = measured_strings_reply( & address, 1 );
			}
			rwlock_read_unlock( & netif_globals.lock );
			return ERROR_CODE;
	}
	return specific_message( callid, call, answer, answer_count );
}

int netif_start_module( async_client_conn_t client_connection ){
	ERROR_DECLARE;

	async_set_client_connection( client_connection );
	netif_globals.networking_phone = connect_to_service( SERVICE_NETWORKING );
	device_map_initialize( & netif_globals.device_map );
	ERROR_PROPAGATE( pm_init());
	rwlock_initialize( & netif_globals.lock );
	if( ERROR_OCCURRED( initialize())){
		pm_destroy();
		return ERROR_CODE;
	}

	async_manager();

	pm_destroy();
	return EOK;
}

void netif_pq_release( packet_id_t packet_id ){
	pq_release( netif_globals.networking_phone, packet_id );
}

packet_t netif_packet_get_1( size_t content ){
	return packet_get_1( netif_globals.networking_phone, content );
}

/** @}
 */
